
#-----------------------------------------------------------------------------
#  Estimate & CalcForecasting

tolEstimate <- function(output, inputs=list(), ..., coeffs=list(), 
  dif=1, AR=list(), MA=list(), period=1, power=1, shift=0) 
{
  # Checks
  stopifnot(inherits(output, "Serie"))
  stopifnot(is.list(inputs))
  for(input in inputs) stopifnot(inherits(input, "Serie"))
  coeffs <- as.list(coeffs)
  stopifnot(is.list(coeffs))
  stopifnot(length(coeffs)==length(inputs) | length(coeffs)==0)
  dif <- as.Polyn(dif)
  stopifnot(inherits(dif, "Polyn"))
  stopifnot(is.list(AR))
  stopifnot(is.list(MA))
  stopifnot(length(AR)==length(MA) | length(AR)==0 | length(MA)==0)
  # Inputs
  setinput <- list()
  if(length(inputs)>0) for(i in 1:length(inputs)) {
    coeff <- if(length(coeffs)>0) coeffs[[i]] else as.Polyn(0.1)
    setinput[[i]] <- tolStruct("@InputDef", coeff, inputs[[i]])
    nme <- if(is.null(names(inputs))) paste("X_", i, sep="")
      else if(names(inputs)[[i]]=="") paste("X_", i, sep="")
      else names(inputs)[[i]]
    names(setinput[[i]]) <- c("", nme)
  }
  # ARIMA
  if(length(AR)==0 & length(MA)==0) {
    AR <- list(as.Polyn(1))
    MA <- list(as.Polyn(1))
  } else if(length(AR)==0) {
    AR <- rep(list(as.Polyn(1)), length(MA))
    MA <- lapply(MA, as.Polyn)
  } else if(length(MA)==0) {
    AR <- lapply(AR, as.Polyn)
    MA <- rep(list(as.Polyn(1)), length(AR))
  } else {
    AR <- lapply(AR, as.Polyn)
    MA <- lapply(MA, as.Polyn)
  }
  modeldef <- tolStruct("@ModelDef", output, power, shift, period, 1, 
    dif, AR, MA, setinput, list())
  ans <- tolExecute("Estimate", modeldef)
  class(ans) <- ".TolEstimate"
  ans
}

print..TolEstimate <- function(x, ...) {
  cat("Coefficients:\n")
  print(coef..TolEstimate(x))
  invisible(x)
}

plot..TolEstimate <- function(x, ...) {
  plot.Serie(list(Original=x[[2]][[1]], 
                  Fitted=fitted..TolEstimate(x)), ...)
  invisible(x)
}

coef.names..TolEstimate <- function(object, ...) {
  pars <- object[[4]]
  nmes <- character(0)
  for(i in 1:length(pars)) {
    nme <- pars[[i]][[1]]
    nme <- gsub("Regular", "", nme)
    nme <- gsub("Estacional ", "S", nme)
    nme <- gsub("Seasonal ", "S", nme)
    nme <- gsub("[(]", "", nme)
    nme <- gsub("[)]", ".", nme)
    nme <- paste(nme, pars[[i]][[3]], sep=".")
    #nme <- pars[[i]][[5]] stderr
    if(!is.na(charmatch("X_", nme))) 
      nme <- paste("(Input ",substr(nme,3,nchar(nme)),")",sep="")
    nmes <- c(nmes, nme)
  }
  nmes
}

coef..TolEstimate <- function(object, ...) {
  # pars <- object[[4]]
  # coefs <- numeric(0)
  # for(i in 1:length(pars)) coefs <- c(coefs, pars[[i]][[4]])
  coefs <- as.numeric(lapply(object[[4]], function(x) x[[4]]))
  names(coefs) <- coef.names..TolEstimate(object)
  coefs
}

residuals..TolEstimate <- function(object, ...) {
  # Obtained residuals according to: 
  # residuals = observations - fitted
  object[[2]][[1]]-fitted..TolEstimate(object)
}

fitted..TolEstimate <- function(object, ...) {
  # Fitted values relative to the observations.
  # Power transformation is reverted.
  power <- object[[2]][[2]]
  shift <- object[[2]][[3]]  
  if(power==0) exp(object[[3]][[7]])-shift
  else (object[[3]][[7]])^(1/power)-shift
}

vcov..TolEstimate <- function(object, ...) {
  vc <- object[[5]][[7]]
  nmes <- coef.names..TolEstimate(object)
  attr(vc, "dimnames") <- list(nmes, nmes)
  vc
}

#-----------------------------------------------------------------------------

tolCalcForecasting <- function(estimation, forecastEnd, ..., alpha=0.05) {
  stopifnot(inherits(estimation, ".TolEstimate"))
  stopifnot(!missing(forecastEnd))
  output <- estimation[[2]][[1]]
  length.out <- if(is.numeric(forecastEnd)) {
    round(forecastEnd)
  } else if(inherits(forecastEnd, c("Date", "POSIXt"))) {
    Ddiff(Slast(output), forecastEnd, Sdating(output))
  }
  cfor <- tolExecute("CalcForecasting", 
    list(estimation[[1]], estimation[[2]], estimation[[3]]), 
    Sfirst(output), Slast(output), length.out, alpha)
  class(cfor) <- ".TolCalcForecasting"
  cfor
}

print..TolCalcForecasting <- function(x, ...) {
  original <- x[[1]]
  forecast <- x[[2]]
  cat(paste("Original from ", as.character(Sfirst(original)), 
            " to ", as.character(Slast(original)), "\n", sep=""))
  cat(paste("Forecast from ", as.character(Sfirst(forecast)), 
    " to ", as.character(Slast(forecast)), "\n", sep=""))
  invisible(x)
}

plot..TolCalcForecasting <- function(x, actual=NULL, ...) {
  series <- c(x[1:4], list(actual))
  names(series) <- c("Original", "Forecast", "Superior", "Inferior", "Actual")
  plot.Serie(series, ...)
  invisible(x)
}

#-----------------------------------------------------------------------------